// dss_other.cpp
// Implements the existing suspension-oblivious and suspension-aware analysis approaches

#include "dss_other.h"
#include "rta.h"
#include "models.h"
#include <algorithm>

// Suspension-oblivious analysis for the case of implicit deadlines
bool dss_other::suspension_oblivious_analysis(const std::vector<DssTask> task_set)
{
	double system_utilization = 0.0;
	for (const auto& task : task_set) {
		system_utilization += static_cast<double>(task.get_wcet() + task.get_suspension()) / static_cast<double>(task.get_period());
		if (system_utilization > 1.0) {
			return false;
		}
	}

	return true;
}

// Suspension-oblivious analysis for the case of constrained deadlines
bool dss_other::suspension_oblivious_rta(const std::vector<DssTask> task_set)
{
	std::vector<SeqTask> seq_task_set(task_set.size());

	for (size_t j = 0; j < task_set.size(); j++) {
		seq_task_set[j].set_deadline(task_set[j].get_deadline());
		seq_task_set[j].set_period(task_set[j].get_period());
		seq_task_set[j].set_wcet(task_set[j].get_wcet() + task_set[j].get_suspension());
		seq_task_set[j].set_jitter(0);
	}

	return rta::analyze_task_set(seq_task_set);
}

static int div_ceil(const int dividend, const int divisor) {
	//double quotient = static_cast<double>(dividend) / static_cast<double>(divisor);
	//return static_cast<int>(std::ceil(quotient));
	return (dividend / divisor) + ((dividend % divisor) != 0);
}

static int div_floor(const int dividend, const int divisor) {
	//double quotient = static_cast<double>(dividend) / static_cast<double>(divisor);
	//return static_cast<int>(std::floor(quotient));
	return (dividend / divisor);
}

struct increasing_period {
	bool operator() (const DssTask& task_a, const DssTask& task_b)
	{
		return (task_a.get_period() < task_b.get_period());
	}
};

// Suspension-aware analysis by Guenzel et al. ("Suspension-aware earliest-deadline-first scheduling analysis", TCAD, 2020)
bool dss_other::suspension_aware_rta(const std::vector<DssTask> task_set)
{
	// sort tasks by increasing period
	std::vector<DssTask> tasks = task_set;
	std::stable_sort(tasks.begin(), tasks.end(), increasing_period());

	int n = tasks.size();
	std::vector<int> R(n);
	for (int i = 0; i < n; i++) {
		R[i] = 0;
	}
	
	for (int k = (n - 1); k >= 0; k--) {
		int T_k = tasks[k].get_period();
		int C_k = tasks[k].get_wcet();
		int S_k = tasks[k].get_suspension();

		// compute carry-in interference
		std::vector<int> A_k(n);
		for (int i = 0; i < n; i++) {
			A_k[i] = 0;
		}

		for (int i = 0; i < n; i++) {
			if (i != k) {
				int T_i = tasks[i].get_period();

				if (i < k) {
					A_k[i] = T_k - (div_floor(T_k, T_i) * T_i);
				}
				if (i > k) {
					A_k[i] = T_k + R[i] - ((div_floor(T_k, T_i) + 1) * T_i);
				}
			}
		}

		// compute response time bounds for task k with different thresholds
		std::vector<int> R_k(n);
		for (int i = 0; i < n; i++) {
			R_k[i] = 0;
		}

		for (int j = 0; j < n; j++) {
			if (j != k) {
				int m_jk = std::max(A_k[j], 0);

				R_k[j] = C_k + S_k + m_jk;
				for (int i = 0; i < n; i++) {
					if (i != k) {
						int T_i = tasks[i].get_period();
						int C_i = tasks[i].get_wcet();
						
						if (A_k[i] <= A_k[j]) {
							int min_a = div_floor(T_k, T_i);
							int min_b = div_ceil((T_k - m_jk), T_i);
							R_k[j] += std::min(min_a, min_b) * C_i;
						}
						else {
							int min_a = div_floor(T_k, T_i) + 1;
							int min_b = div_ceil((T_k - m_jk), T_i);
							R_k[j] += std::min(min_a, min_b) * C_i;
						}
					}
				}
			}
		}
		
		// store response time without threshold in unused location k
		R_k[k] = C_k + S_k;
		for (int i = 0; i < n; i++) {
			if (i != k) {
				int T_i = tasks[i].get_period();
				int C_i = tasks[i].get_wcet();
				
				R_k[k] += (div_floor(T_k, T_i) + 1) * C_i;
			}
		}

		// look for the minimum value
		int R_k_min = R_k[0];
		for (int i = 1; i < n; i++) {
			if (R_k_min > R_k[i]) {
				R_k_min = R_k[i];
			}
		}
		R[k] = R_k_min;

		// check schedulability
		if (R[k] > T_k) {
			return false;
		}
	}

	return true;
}
