// experiments.cpp
// Implements systematic experiments to evaluate the schedulability ratio obtained with different analyses on randomly generated task sets

#include "experiments.h"
#include "models.h"
#include "dss_rta.h"
#include "dss_other.h"
#include <vector>
#include <random>
#include <iostream>
#include <future>

std::random_device rd;
std::mt19937 rng(rd());

double draw_uniform_distribution_real(double min, double max)
{
	if (min > max) {
		std::cerr << "Error: invalid distribution range" << std::endl;
	}
	std::uniform_real_distribution<double> distribution(min, max);
	return distribution(rng);
}

int draw_uniform_distribution_integer(int min, int max)
{
	if (min > max) {
		std::cerr << "Error: invalid distribution range" << std::endl;
	}
	std::uniform_int_distribution<int> distribution(min, max);
	return distribution(rng);
}

int draw_log_uniform_distribution_integer(int min, int max)
{
	if (min > max) {
		std::cerr << "Error: invalid distribution range" << std::endl;
	}
	int granularity = 1;

	std::uniform_real_distribution<double> distribution(log(min), log(max + granularity));
	double floor = std::floor(exp(distribution(rng)) / static_cast<double>(granularity));
	int value = static_cast<int>(floor) * granularity;
	if (value < min) {
		value = min;
		std::cerr << "Error: invalid distribution range" << std::endl;
	}
	if (value > max) {
		value = max;
		std::cerr << "Error: invalid distribution range" << std::endl;
	}

	return value;
}

// Implements the UUniFast algorithm by Bini and Buttazzo ("Measuring the performance of schedulability tests", Real-Time Systems, 2005)
std::vector<double> uunifast(int num_tasks, double total_utilization)
{
	std::vector<double> utilizations(num_tasks);
	if (num_tasks < 1) {
		return utilizations;
	}

	double util_sum = total_utilization;
	for (int i = 0; i < num_tasks - 1; i++) {
		double base = draw_uniform_distribution_real(0.0, 1.0);
		double exponent = static_cast<double>(static_cast<double>(1.0) / static_cast<double>(num_tasks - i - 1)); 
		double next_util_sum = util_sum * pow(base, exponent);
		utilizations[i] = util_sum - next_util_sum;
		util_sum = next_util_sum;
	}
	utilizations[num_tasks - 1] = util_sum;

	return utilizations;
}

DssTask generate_task(double utilization, int min_period, int max_period, double min_susp_factor, double max_susp_factor, double deadline_factor)
{
	DssTask task;
	
	if ((utilization > 1.0) || (min_period > max_period) || (min_susp_factor > 1.0) || (max_susp_factor > 1.0) || (min_susp_factor > max_susp_factor) || (deadline_factor > 1.0)) {
		std::cerr << "Error: invalid parameter range" << std::endl;
		return task;
	}

	int period = draw_log_uniform_distribution_integer(min_period, max_period);
	task.set_period(period);

	int wcet = static_cast<int>(static_cast<double>(period) * utilization);
	task.set_wcet(wcet);

	double min_suspension = static_cast<double>(period - wcet) * min_susp_factor;
	double max_suspension = static_cast<double>(period - wcet) * max_susp_factor;
	int suspension = static_cast<int>(draw_uniform_distribution_real(min_suspension, max_suspension));
	task.set_suspension(suspension);

	if (deadline_factor == 1.0) {
		task.set_deadline(period);
	}
	else {
		int slack = period - wcet;
		double min_deadline = static_cast<double>(wcet) + (static_cast<double>(slack) * deadline_factor);
		double max_deadline = static_cast<double>(period);
		int deadline = static_cast<int>(draw_uniform_distribution_real(min_deadline, max_deadline));
		task.set_deadline(deadline);
	}

	task.set_wcrt_ub(0);

	return task;
}

std::vector<DssTask> generate_task_set(double utilization, int num_tasks, int min_period, int max_period, double min_susp_factor, double max_susp_factor, double deadline_factor)
{
	std::vector<DssTask> task_set(num_tasks);
	std::vector<double> utilizations = uunifast(num_tasks, utilization);
	for (int i = 0; i < num_tasks; i++) {
		task_set[i] = generate_task(utilizations[i], min_period, max_period, min_susp_factor, max_susp_factor, deadline_factor);
	}

	return task_set;
}

struct IterationResults {
	bool schedulable_rta = false; // task set schedulable according to the proposed analysis
	bool schedulable_obl = false; // task set schedulable according to the existing suspension-oblivious analysis
	bool schedulable_oth = false; // task set schedulable according to the existing suspension-aware analysis
	bool schedulable_or = false; // task set schedulable according to either the proposed analysis or the existing suspension-aware analysis
};

IterationResults run_iteration(int i, double utilization, int num_tasks, int min_period, int max_period, double min_susp_factor, double max_susp_factor, double deadline_factor)
{
	IterationResults results;

	std::vector<DssTask> task_set = generate_task_set(utilization, num_tasks, min_period, max_period, min_susp_factor, max_susp_factor, deadline_factor);

	bool constrained_deadlines = false;
	for (const auto& task : task_set) {
		if (task.get_deadline() != task.get_period()) {
			constrained_deadlines = true;
		}
	}
	if (constrained_deadlines) {
		results.schedulable_obl = dss_other::suspension_oblivious_rta(task_set);
		results.schedulable_oth = false;
	}
	else {
		results.schedulable_obl = dss_other::suspension_oblivious_analysis(task_set);
		results.schedulable_oth = dss_other::suspension_aware_rta(task_set);
	}

	results.schedulable_rta = dss_rta::analyze_task_set(task_set, true);
	results.schedulable_or = (results.schedulable_rta || results.schedulable_oth);

	//std::cout << ".";
	return results;
}

// Performs the experiment based on the given parameters
void experiments::sweep_utilization(ExperimentParameters exp_parameters)
{
	// processing parameters
	bool parallel_mode = exp_parameters.parallel_mode;

	// experiment parameters
	double min_utilization = exp_parameters.min_utilization;
	double max_utilization = exp_parameters.max_utilization;
	double utilization_step = exp_parameters.utilization_step;
	int num_iterations = exp_parameters.num_iterations;

	// generation parameters
	int num_tasks = exp_parameters.num_tasks;
	int min_period = exp_parameters.min_period;
	int max_period = exp_parameters.max_period;
	double min_susp_factor = exp_parameters.min_susp_factor;
	double max_susp_factor = exp_parameters.max_susp_factor;
	double deadline_factor = exp_parameters.deadline_factor;
	
	// print parameters
	std::cout << "Experiment parameters:" << std::endl;
	std::cout << "pmode: " << exp_parameters.parallel_mode << std::endl;
	std::cout << "umin: " << exp_parameters.min_utilization << std::endl;
	std::cout << "umax: " << exp_parameters.max_utilization << std::endl;
	std::cout << "ustep: " << exp_parameters.utilization_step << std::endl;
	std::cout << "niter: " << exp_parameters.num_iterations << std::endl;
	std::cout << "ntasks: " << exp_parameters.num_tasks << std::endl;
	std::cout << "tmin: " << exp_parameters.min_period << std::endl;
	std::cout << "tmax: " << exp_parameters.max_period << std::endl;
	std::cout << "smin: " << exp_parameters.min_susp_factor << std::endl;
	std::cout << "smax: " << exp_parameters.max_susp_factor << std::endl;
	std::cout << "dlfac: " << exp_parameters.deadline_factor << std::endl;
	std::cout << std::endl;

	// generate utilization vector
	int num_steps = 0;
	std::vector<double> utilization_values(0);
	double u = min_utilization;
	while (u < max_utilization) {
		u = min_utilization + static_cast<double>(num_steps) * utilization_step;
		utilization_values.push_back(u);
		num_steps++;
	}
	std::vector<int> schedulable_rta(num_steps);
	std::vector<int> schedulable_obl(num_steps);
	std::vector<int> schedulable_oth(num_steps);
	std::vector<int> schedulable_or(num_steps);

	// run experiments
	for (int u = 0; u < num_steps; u++) {
		double utilization = utilization_values[u];
		std::cout << "Step " << u+1 << "/" << num_steps << ", util " << utilization << std::endl;

		if (parallel_mode) {
			std::vector<std::future<IterationResults> > future_results;
			for (int i = 0; i < num_iterations; i++) {
				future_results.push_back(std::async(run_iteration, i, utilization, num_tasks, min_period, max_period, min_susp_factor, max_susp_factor, deadline_factor));
			}

			for (auto& future : future_results) {
				IterationResults iteration_results = future.get();
				if (iteration_results.schedulable_rta) {
					schedulable_rta[u]++;
				}
				if (iteration_results.schedulable_obl) {
					schedulable_obl[u]++;
				}
				if (iteration_results.schedulable_oth) {
					schedulable_oth[u]++;
				}
				if (iteration_results.schedulable_or) {
					schedulable_or[u]++;
				}
			}
		}
		else {
			for (int i = 0; i < num_iterations; i++) {
				IterationResults iteration_results = run_iteration(i, utilization, num_tasks, min_period, max_period, min_susp_factor, max_susp_factor, deadline_factor);
				if (iteration_results.schedulable_rta) {
					schedulable_rta[u]++;
				}
				if (iteration_results.schedulable_obl) {
					schedulable_obl[u]++;
				}
				if (iteration_results.schedulable_oth) {
					schedulable_oth[u]++;
				}
				if (iteration_results.schedulable_or) {
					schedulable_or[u]++;
				}
			}
		}

		//std::cout << std::endl;
	}

	// get schedulability ratios
	std::vector<double> sched_ratios_rta(num_steps);
	std::vector<double> sched_ratios_obl(num_steps);
	std::vector<double> sched_ratios_oth(num_steps);
	std::vector<double> sched_ratios_or(num_steps);
	for (int u = 0; u < num_steps; u++) {
		sched_ratios_rta[u] = static_cast<double>(schedulable_rta[u]) / static_cast<double>(num_iterations);
		sched_ratios_obl[u] = static_cast<double>(schedulable_obl[u]) / static_cast<double>(num_iterations);
		sched_ratios_oth[u] = static_cast<double>(schedulable_oth[u]) / static_cast<double>(num_iterations);
		sched_ratios_or[u] = static_cast<double>(schedulable_or[u]) / static_cast<double>(num_iterations);
	}

	// print results
	std::cout << std::endl;
	std::cout << "Results (utilization, schedulability)" << std::endl;
	std::cout << "util\t\trta\t\tobl\t\toth\t\tor" << std::endl;
	for (int u = 0; u < num_steps; u++) {
		//std::cout << utilization_values[u] << "\t\t" << sched_ratios_rta[u] << "\t\t" << sched_ratios_obl[u] << "\t\t" << sched_ratios_oth[u] << "\t\t" << sched_ratios_or[u] << std::endl;
		printf("%-15g ", utilization_values[u]);
		printf("%-15g ", sched_ratios_rta[u]);
		printf("%-15g ", sched_ratios_obl[u]);
		printf("%-15g ", sched_ratios_oth[u]);
		printf("%-15g ", sched_ratios_or[u]);
		printf("\n");
	}
}

// Runs unit tests for this module
bool experiments::test()
{
	for (int i = 0; i < 1000; i++) {
		std::vector<double> utilizations = uunifast(3, 1.0);
		double total_util = 0.0;
		for (const auto util : utilizations) {
			total_util += util;
		}
		if (std::abs(total_util - 1.0) > 0.000001) {
			return false;
		}
	}
	for (int i = 0; i < 1000; i++) {
		std::vector<double> utilizations = uunifast(4, 1.0);
		double total_util = 0.0;
		for (const auto util : utilizations) {
			total_util += util;
		}
		if (std::abs(total_util - 1.0) > 0.000001) {
			return false;
		}
	}
	for (int i = 0; i < 1000; i++) {
		std::vector<double> utilizations = uunifast(10, 1.0);
		double total_util = 0.0;
		for (const auto util : utilizations) {
			total_util += util;
		}
		if (std::abs(total_util - 1.0) > 0.000001) {
			return false;
		}
	}

	return true;
}
